/* Help File Basic Structures and Procedures.
   Copyright 1994, 1995 Bent Lynggaard.
*/

/* 1994-08-13 - v. 1.00 - Bent Lynggaard.
   1994-08-17 - v. 1.01 - Edited WINDOWSTRUCT
   1994-08-20 - v. 1.02 - New WINDOWSTRUCT30 (HC30)
   1994-08-23 - v. 1.03 - Edited structures, new structures
   1994-08-28 - v. 1.04 - Ditto, and structures removed
   1994-09-13 - v. 1.05 - New structures and functions
   1994-10-23 - v. 1.06 - Use of file handles rather than streams
   1995-02-24 - v. 1.07 - nextInfo format for data only, CONTEXTDATA,
						  BLOCKTABLE field names changed from resNN to wHH
   1995-04-11 - v. 1.08 - Function getLatestInfoDataOffset, comments edited
   1995-04-14 - v. 1.09 - findUnused and linkUnused functions
   1995-04-15 - v. 1.10 - library instead of obj, function writeBlock
   1995-04-18 - v. 1.11 - Function findIndexTableEntry
*/

#ifdef __cplusplus
extern "C" {            /* Assume C declarations for C++ */
#endif

/* errors returned by Help Base functions: */

#define HBERR_PAST_END 1
#define HBERR_SHORT 2
#define HBERR_NOTOPEN 3

#define HBERR_OPEN 8
#define HBERR_MARKER 9
#define HBERR_VERSION 10
#define HBERR_LENGTH 11
#define HBERR_CANT_ALLOC 12
#define HBERR_TABLE_LIMIT 13
#define HBERR_NOT_INIT 14
#define HBERR_NO_ENTRY 15
#define HBERR_BAD_FORMAT 16

#define HBERRORS 200
/* it is suggested to use HBERRORS + a HBERR_xxx code as error return codes */

#ifndef __WINDOWS_H
typedef unsigned char	BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;
typedef long			LONG;
typedef int				BOOL;
typedef int				HFILE;	/* file handle */

typedef DWORD COLORREF;
#define RGB(r,g,b) ((COLORREF)(((BYTE)(r)|((WORD)(g)<<8))|(((DWORD)(BYTE)(b))<<16)))
#define GetRValue(rgb) ((BYTE)(rgb))
#define GetGValue(rgb) ((BYTE)(((WORD)(rgb)) >> 8))
#define GetBValue(rgb) ((BYTE)((rgb)>>16))
#endif // __WINDOWS_H

typedef struct {
  BYTE red;
  BYTE green;
  BYTE blue;
  BYTE flag;
} COLORSTRUCT;

typedef union { /* for easier access to colors */
  COLORREF		cr;
  COLORSTRUCT	cs;
} COLORUNION;

/* The first 16 bytes of every WINHELP file contains information that
   identifies the file as a WINHELP file, and defines an offset for an
   index, and the length of the file.
*/

typedef struct {
	WORD marker;	/* '?_' */
	char ver;		/* 3 (version?) */
	char subver;	/* 0 */
	DWORD index;	/* offset of index block */
	DWORD unused;	/* link to unused sections */
	/* Field "unused" is either (DWORD)-1 or the offset of the first block
	   in a linked list of unused sections. The offset of the next unused
	   block is in field "datalength" of the BLOCKINFO part of the current
	   block, or (DWORD)-1 in the last unused block in the chain.
	*/
	/* When the index grows to the limit allocated for it in the file, a
	   new and larger block is allocated. The previous block is linked onto
	   the unused blocks chain, and index is assigned the new offset.
	   Unused blocks are reused when a new section (normally a bitmap)
	   will fit. If no new section will fit for the rest of the file, field
	   "unused" points to the unused sections chain.
	*/
    DWORD length;		/* of file */
} WINHELPHEADER;

#define WHMARKER ('?' + ('_'<<8))

/* The rest of the file is divided into sections. These sections are
   normally in a linked structure (unless the file is optimized for CDROM),
   but the important key to the structure is the index, which contains
   information about the name and location of all other blocks. The offset
   of the index block is in WINHELPHEADER's index field.

   Some help files can have an unused section, see struct WINHELPHEADER.

   All the offsets are to a 9 byte part of the file. The first two "long"s
   are the size of the block (thus also normally the offset to the next
   block) and normally the length of the data (i.e. 9 less than the length
   of the block). The next byte is 04 for the index block and 00 for all
   other blocks in the cases observed.

   The first word of the data can be 1) a word identifying the data, 2)
   the number of entries in the following data field, or 3) part of the
   data field.

   The following data structures have been observed:
     The unnamed index structure.
     An unnamed discarted index structure, see struct WINHELPHEADER.
     filename	data included with a [BAGGAGE] file name.
     |CONTEXT	pairs of context string hash values and offsets.
     |CTOXMAP	?
     |FONT		Fonts known to this help file.
     |KWBTREE	All the search keywords.
     |KWDATA	A link between the keywords and the topic titles.
     |KWMAP		?
     |Phrases	Compression phrases table.
     |SYSTEM	Window title, colors, macros, etc.
     |TOPIC	All the topics in the file.
     |TTLBTREE	Title table.
     |bm0	\
     |bm1	 > Bitmaps (HC30 uses bm0 bm1 ... (without a bar))
     ...	/
*/

typedef struct {
	/* The first 11 bytes are read from the file at specified offsets */
	DWORD length;		/* of link + data */
		/* normally the offset to next link, but not necessarely if option
		   OPTCDROM is on.
		*/
	DWORD datalength;	/* normally, but not always, length-9 */
	BYTE flag;		/* 4 for index flag, 0 for others */
	WORD info;		/* may or may not be meaningful */
	/* The name and offset fields can be read from the file's index block.
		 They are filled out first, and then the offset field is used to find
		 the information for the first four fields.
	*/
	char name[13];	/* allow a file name */
	DWORD offset;		/* to start of link */
} BLOCKINFO;

/* Some data structures have a 38 bytes header followed by a variable
   number of fixed length data blocks.

   The structures BLOCKTABLE and BLOCKUNIT are used to access these blocks.
   The first 38 (decimal) bytes of a BLOCKTABLE is read from the file, and
   the rest is initialized under program control. The BLOCKUNIT is the first
   8 bytes of every fixed length data block.

   The data following the BLOCKUNIT are pairs of variable length strings
   and fixed length data (or vica versa: data/string). BLOCKUNITS are linked
   with forwards and backwards links. Some block types are normally linked
   in contiguous order, whereas others can be linked in strange patterns.

   The functions initInfoBlock() and nextInfo() are used to access the
   information.

   The following structures use this format:
     The index:	String/4 byte offset.
     |CONTEXT:	4 byte hash value, 4 byte offset.
     |KWBTREE:	String/2 byte count/4 byte offset.
     |TTLBTREE:	4 byte offset/string.

   The hash value for context strings is calculated from the character
   values, starting with a hash value = 0:
     Char		Hex		Dec
				00		0		Not used
     1-9		01-09	1-9
     0			0A		10
				0B		11		Not used
     .			0C		12
     _			0D		13
				0E-10	14-16	Not used
     A-Z, a-z	11-2A	17-42
   New hash value = (old hash value * 0x2B + new character) & 0xFFFFFFFF.
   Thus "aaaaaaaaa" and "gv08_i" both give the hash value 0x290C30DC.

   The function readDataBlock is for internal use, but it can be used
   to check the link sequence:
     BLOCKTABLE mytable;
     ...
     initInfoBlock(<file>, <offset>, &mytable, <dummy>, <dummy>, <dummy>);
     while (mytable.curTable->next != 0xFFFF) {
       ...
       readDataBlock(&mytable, mytable.curTable->next);
       ...
     }
*/

typedef struct {
	WORD unused;		/* unused bytes at end */
	WORD localEntries;	/* entries in this block */
	WORD prev;			/* backward link (block #, or FFFF) */
	WORD next;			/* forward link */
} BLOCKUNIT; /* initial table in file blocks with tabular info */

typedef struct {
	/* the first 38 bytes are read from the help file */
	WORD marker;		/* 0x293B */
	BYTE flag;			/* same as flag after link, 00 or 04 */
	BYTE zero;			/* 00 */
	WORD blockLength;	/* 0400 or 0800 */
	WORD type;			/* 347A, 3269, 7A4C, 344C */
	WORD w08;
	WORD w0A;
	WORD w0C;
	WORD w0E;
	WORD w10;
	WORD w12;
	WORD w14;
	WORD w16;			/* often 0 */
	WORD w18;			/* often w1A - 1 */
	WORD w1A;			/* often blocks - 1 */
	WORD w1C;			/* often FFFF */
	WORD blocks;		/* number of blocks */
	WORD w20;
	DWORD entries;		/* number of entries */

	/* the rest of the structure keeps track of reading */
	HFILE hFile;		/* file handle */
	DWORD offset;		/* in file to first subblock */
	BLOCKUNIT *curTable;	/* pointer to 'blocklength' bytes */
	DWORD curEntry;		/* current entry number */
	char *curP;			/* initially, (char*)curTable + 8 */
	WORD curBlock;		/* current block */
	WORD curSub;		/* entry in current block */
	WORD stringBufLength;	/* length of string buffer for returned
				 string, max length = stringbuf - 1 */
	WORD dataLength;	/* length of data returned */
		/* data are stored as alternating variable length 0-terminated
		   strings and fixed length data.
		*/
	int stringFirst;	/* 0: data/string; 1: string/data. */
} BLOCKTABLE; /* of "index", "KWBTREE", "TTLBTREE", and "CONTEXT" blocks */

typedef struct {
	BYTE b08;			/* 06 */
	BYTE compr;			/* compression mode 0, 1, or 2 */
	WORD w0A;			/* 00C0 */
	WORD w0C;			/* 00C0 */
	BYTE b0E;			/* 02 (maybe 2 * bitplanes) */
	BYTE bitsX2;		/* 2 * bits/pixel */
	WORD widthX2;		/* 2 * width */
	WORD heightX2;		/* 2 * height */
	WORD colorsX2;		/* 2 * colors */
	WORD colors2X2;		/* ditto or 0 */
	WORD bytesX2;		/* 2 * bites in bitmap */
	WORD w1A;
	WORD length;		/* of this structure + palette table */
	WORD w1E;
	WORD w20;
	WORD w22;
} BMPINFO;

typedef struct {
	WORD type;			/* 506C */
	WORD w02;			/* 0001 */
	DWORD d04;			/* 00000008 */
	BMPINFO info;
} BMPTABLE;

typedef struct {
	DWORD hash;
	DWORD offs;
} CONTEXTDATA;

typedef struct {
	WORD n;
	DWORD offset;
} KWBTREEDATA;

typedef struct {
	DWORD offset;
	WORD n;
} CTOXMAPDATA;


/* The |SYSTEM structure defines the help window(s). The first 12 byte are
   identical for HC30 and HC31 compiled files except that byte 2 is 0x0F
   and 0x15 respectively, and the flag in bytes 10/11 have different
   meanings.

   HC30 compilations have a fixed length (33 bytes) title field next.

   HC31 compilations have linked entries of data recording information in
   the [OPTION], [WINDOWS], and [CONFIG] sections of the help project
   file. The link has a 2 byte type and a 2 byte length of the following
   data. Thus the next link is offset (length + 4) bytes from the current
   position. The length of the |SYSTEM structure defines the end of the
   linked list.
*/

typedef struct {
	WORD type;	/* 0x036C */
	BYTE c1;	/* 0x0F (HC30) or 0x15 (HC31) */
	BYTE c2;	/* 0 */
	BYTE c3;	/* 1 */
	BYTE c4;	/* 0 */
	long time;	/* seconds since 1980-01-01 */
	WORD flag;	/* bit 2 = compressed if HC31 */
} SYSTEM;

typedef struct { /* between local blocks of info in SYSTEM block */
	WORD type;		/* 1 - 6, 8 observed */
	WORD length;	/* of following entry */
		/* next link is at current position + 4 + length */
} LOCALLINK;

/* info types */
#define TITLE_TYPE 1
	/* 'length' bytes 0-terminated string, max 51, [OPTIONS] */
#define COPYRIGHT_TYPE 2
	/* 'length' bytes 0-terminated string, max 51, [OPTIONS] */
#define CONTENTS_TYPE 3
	/* 4 bytes offset to Contents topic, [OPTIONS] */
#define MACRO_TYPE 4
	/* 'length' bytes 0-terminated string, max 513 (?), [CONFIG] */
#define ICON_TYPE 5
	/* 'length' (?) bytes icon bitmap, [OPTIONS] */
#define WINDOW_TYPE 6
	/* 90 bytes windows structure, [WINDOWS] */
/* type 7 has not been observed */
#define CITATION_TYPE 8
	/* 'length' bytes 0-terminated string, max 2000 (?), [OPTIONS] */

typedef struct { /* HC31 window structure */
	WORD flag;
		/* bit   0:
		   bit   1:
		   bit   2: 1 if title is specified, else 0.
		   bits  6-3: 1111b if window position/size is defined, else 0000b
		   bit   7:
		   bit   8: 1 if scrolling color defined, else 0.
		   bit   9: 1 if nonscrolling (header) color defined, else 0.
		   bit  10: 1 if window always on top, else 0.
		   bit  11:
		   bit  12:
		   bit  13:
		   bit  14:
		   bit  15:
		*/
	char kind[10];		/* "main" or "secondary" */
	char name[9];
	char title[51];		/* if flag & WS_TITLE */
	WORD xCoord;		/* position, if flag & WS_POSSIZE */
	WORD yCoord;
	WORD width;			/* size, ditto */
	WORD height;
	WORD state;			/* 0 = normal, 1 = maximized */
	COLORREF scrollColor;		/* if flag & WS_SCOLOR */
	COLORREF nonscrollColor;	/* if flag & WS_NSCOLOR */
} WINDOWSTRUCT;

#define WS_TITLE 0x004
#define WS_POSSIZE 0x0078
#define WS_SCOLOR 0x0100
#define WS_NSCOLOR 0x0200
#define WS_ONTOP 0x0400

int contextComp(const CONTEXTDATA *a, const CONTEXTDATA *b);
/* Compares two CONTEXTDATA structures with .offs as the primary criterion
   and .hash as a secondary criterion.
   Returns:
     0 if both are equal.
     0x7FFF if the structures compare positive.
     0xFFFF if the structures compare negative.
   Compatible with qsort.
*/

void contextSortHUGE(CONTEXTDATA huge *context, WORD size);
/* Sorts a CONTEXTDATA structure with .offs as the primary criterion and .hash
   as a secondary criterion. Up to 65,535 elements are accepted.
*/

#define NOWRITE 0
#define WRITEACCESS 1

#ifdef __cplusplus
int openHelpFile(HFILE *hFile, char *name, WINHELPHEADER *root = NULL,
	int writeaccess = NOWRITE);
#else
int openHelpFile(HFILE *hFile, char *name, WINHELPHEADER *root,
	int writeaccess);
#endif
/* Opens file 'name' with handle 'hFile' for access defined by parameter
   writeaccess, and reads the header to *root if root != NULL. If writeaccess
   == 0, the file is opened for read access only, else for read/write access.
   Note that the declaration of the function assigns default values of root
   and writeaccess supported by C++. Use NULL for root if no memory is
   allocated and writeaccess must be specified.
   Returns 0 or
     HBERR_OPEN: file open error;
     HBERR_SHORT: file length < 16 bytes;
     HBERR_MARKER: marker not '?_';
     HBERR_VERSION: vers, subvers not 3, 0;
     HBERR_LENGTH: file length != root->length;
   The file is opened unless HBERR_OPEN is returned.
*/
/* help_opn.c */


int readHelpRoot(HFILE hFile, WINHELPHEADER *root);
/* Reads root record and checks consistency.
   Returns 0 or
     HBERR_SHORT: file length < 16 bytes;
     HBERR_MARKER: marker not '?_';
     HBERR_VERSION: vers, subvers not 3, 0;
     HBERR_LENGTH: file length != root->length;
*/
/* help_opn.c */

int readBlock(HFILE hFile, DWORD offset, void *buffer, WORD size);
/* Reads 'size' bytes from file 'hFile', position 'offset' to 'buffer'.
   Returns 0 or
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: less than size bytes could be read;
*/
/* help_rdb.c */

int writeBlock(HFILE hFile, DWORD offset, void *buffer, WORD size);
/* Writes 'size' bytes to file 'hFile', position 'offset' from 'buffer'.
   Returns 0 or
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: less than size bytes could be written;
*/
/* help_wrb.c */

int readLink(HFILE hFile, DWORD offset, BLOCKINFO *buffer);
/* Reads link (2 longs) + flag (char) + info (short) from file 'hFile', position
   'offset' to 'buffer'.
   Returns
     0: link OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: less than 11 bytes could be read;
*/
/* help_rdl.c */

/* int readTable(HFILE hFile, DWORD offset, BLOCKTABLE *buffer); */
/* Reads BLOCKTABLR from file 'hFile', position 'offset'+11 to 'buffer', and
   leaves the file positioned after the block.
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: less than sizeof(BLOCKTABLE) bytes could be read;
*/

#define STRINGFIRST 1
#define STRINGLAST 0

int readStringData(HFILE hFile, char *str, int maxlength,
  void *block, int blocklength, int stringFirst);
/* Reads a string and a block of data from file 'hFile' (current position).
   Maximum 'maxlength' characters are copied to 'str[maxlength+1]', and
   'blocklength' characters are copied to *'block'. If stringFirst == 0,
   the string is the last of the two, else the string is the first. The
   file is positioned after the string/data, also if the string is longer
   than 'maxlength'.
   If the actual string lenght matters, the user must calculate it from
   the file position before and after the call, and 'blocklength'.
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
*/
/* help_rsd.c */

int readDataBlock(BLOCKTABLE *table, WORD n);
/* Reads block # 'n' of data as defined in 'table', and updates 'table'
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
     HBERR_TABLE_LIMIT: requested to read past end of table;
     HBERR_NOT_INIT: *table not initialized;
*/
/* help_inf.c */

int readInfoBlock(HFILE hFile, DWORD offset, BLOCKTABLE *table);
/* Reads 38 bytes of BLOCKTABLE data from file 'hFile' pos. 'offset'.
   Assigns table->curTable=NULL;
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
*/
/* help_rdi.c */

int initInfoBlock(HFILE hFile, DWORD offset, BLOCKTABLE *table,
  WORD stringBufLength, WORD dataLength, int stringFirst);
/* Reads 38 bytes of BLOCKTABLE data from file 'hFile' pos. 'offset'.
   Allocates memory for a buffer and reads the first block of data.
   Updates 'table'.
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
     HBERR_CANT_ALLOC: insufficient heap space;
*/
/* help_inf.c */

int nextInfo(BLOCKTABLE *table, char *str, void *data);
/* Reads a string and a set of data as defined in 'table', updates 'table'.
   If stringBufLength was defined as 0, only data are read (CONTEXT).
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
     HBERR_TABLE_LIMIT: requested to read past end of table;
     HBERR_NOT_INIT: *table not initialized;
*/
/* help_nin.c */

int readInfoEntry(BLOCKTABLE *table, DWORD entry, char *str, void *data);
/* Reads a string and a set of data from entry 'entry' in 'table', positions
   'table' at next entry.
   If stringBufLength was defined as 0, only data are read (CONTEXT).
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
     HBERR_TABLE_LIMIT: requested to read past end of table;
     HBERR_NOT_INIT: *table not initialized;
*/
/* help_rde.c */

void freeInfoTable(BLOCKTABLE *table);
/* Frees the memory allocated for the buffer and set curTable to NULL
*/
/* help_inf.c */

DWORD getLatestInfoDataOffset(void);
/* Returns the offset in the file of the data part of the latest info
   accessed with nextInfo.
*/
/* help_nin.c */

int readBMPtable(HFILE hFile, DWORD offset, BMPTABLE *bmp);
/* Reads BMPTABLE from file 'hFile' pos. 'offset'.
   Returns
     0: OK;
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
*/
/* help_bmp.c */

int findUnusedBlock(HFILE hFile, DWORD *pLength, DWORD *pOffset);
/* Finds the smallest unused block no less than *pLength and unlinks it,
   assigns its offset to *pOffset, and assigns its actual length to *pLength.
   If no block large enough was found, *pOffset is 0 and *pLenght unchanged.
   Returns 0 or error returned by readHelpRoot, readBlock, or writeBlock.
   Note that the returned value is not very interestering (it is 0 provided
   that hFile represents a help file compiled with HC30 or HC31). The
   indication of success is *pOffset != 0. *pLength and *pOffset are always
   unchanged and 0 if an error is returned.
*/
/* help_fiu.c */

int linkUnusedBlock(HFILE hFile, DWORD offset);
/* Inserts the specified block into the linked chain of unused blocks.
   Returns 0 or error returned by readHelpRoot, readBlock, or writeBlock.
   It is unlikely that the function will return anything but 0.
*/
/* help_liu.c */

int findIndexTableEntry(HFILE hFile, char *name, DWORD *offset);
/* Searches the index defined in root for entry name and assigns the
   entry's offset to offset, or assigns 0 to offset if failure.
   Returns 0 or
     HBERR_PAST_END: offset past end of file;
     HBERR_SHORT: insufficient bytes could be read;
     HBERR_CANT_ALLOC: insufficient heap space;
     HBERR_TABLE_LIMIT: requested to read past end of table;
     	(errors above returned from readLink, initInfoBlock, nextInfo)
     HBERR_BAD_FORMAT: unrecognized index table format;
     HBERR_NO_ENTRY: entry not found in index table;
*/
/* help_fie.c */

#ifdef __cplusplus
}
#endif

